/*
 *	emu.c
 *
 *	Famicom/EMU
 *
 *	* Thu Feb 03 00:00:00 JST 2005 Naoyuki Sawa
 *	- 쐬JnB
 */
#include "app.h"

/****************************************************************************
 *	
 ****************************************************************************/

int
famicom_init(const char* fname)
{
	int i;
	const unsigned char* p;

	/* ͂NA܂B */
	input = input_shift = 0;

	/* CVRAMNA܂B */
	memset(m6502mem, 0, sizeof m6502mem);
	memset(ppu_vram, 0, sizeof ppu_vram);

	/* ROMt@C̃tbVAhX擾܂B */
	stream_stop();  /* ĂȂ͂ǁAÔߊmɒ~ */
	ldirect_free(); /* ߂ȂĂvǁAÔ */
	ines = mmap(fname, NULL, 1/*DEFRAG*/);
	ldirect_init(LDIRECT_INTERVAL);
	if(!ines) goto L_ERR;

	/* wb_mF܂B */
	if(memcmp(ines->nes, "NES", 3) != 0) goto L_ERR;
	if(ines->_1a != '\x1a') goto L_ERR;

	/* PRG-ROMACHR-ROM̐擪AhX߂܂B */
	p = (void*)(ines + 1);
	if(ines->rom_control1 & (1<<2)) {
		p += 512; /* 512-byte trainer present */
	}
	prgrom = (void*)p;
	p += 0x4000 * ines->prgrom_page_count;
	chrrom = (void*)p;

	/* Mapper擾܂B */
	i = (ines->rom_control2 & 0xf0) | (ines->rom_control1 >> 4);
	mapper = mapper_list[i];
	if(!mapper) goto L_ERR;

	/* Mapper܂B */
	if(!mapper->init) DIE();
	if(mapper->init() != 0) goto L_ERR;

	/* CPU,PPU,APUZbg܂B */
	m6502cpu_reset(famicom_read, famicom_write);
	ppu_reset(ines->rom_control1);
	apu_reset(m6502mem);

	return 0;

L_ERR:
	return -1;
}

int
famicom_stream_callback(short* wbuff, int param)
{
	apu_mix(wbuff);
	return 0;
}

/****************************************************************************
 *	
 ****************************************************************************/

int
famicom_run(const char* fname, int sound)
{
	int retval;
	int cnt;
	int cycle;
	int pad;
	unsigned char* draw;

	retval = famicom_init(fname);
	if(retval != 0) {
		return retval;
	}

	if(sound) {
		stream_play(APUBUFLEN, famicom_stream_callback, 0, 0);
	}

	cnt = 0;
	draw = NULL;
	for(;;) {
		cnt += FPS;
		if(cnt >= 60) {
			cnt -= 60;
			draw = vbuff;
		}

		/* 1t[(1/60b)̏B */
		do {
			cycle = ppu_update(draw);
			if(ppu.flags & PPU_FLAGS_INTERRUPT) {
				ppu.flags &= ~PPU_FLAGS_INTERRUPT;
				m6502cpu_nmi();
			}
			m6502cpu_run(cycle / 2);
		} while(ppu.scan_end);

		/* APU͎ԂœĂ̂ŁAIRQ͖{60HzŔ܂B
		 * do`while̒IRQ𔭍sĂ܂ƁAG~[VԊZ60Hz𒴂Ă܂܂B
		 * G~[VԂ60Hzɓ悤Ado`while̊OIRQ𔭍s邱Ƃɂ܂B
		 */
		if(apu.irq) { /* QƂȂ̂Ŕrsv */
			m6502cpu_irq();
		}

		/*  0:A    1:B    2:SEL  3:ST   4:UP  5:DN  6:LF  7:RI
		 *  8:?    9:?   10:?   11:?   12:?  13:?  14:?  15:?
		 * 16:Sig 17:Sig 18:Sig 19:Sig 20:0  21:0  22:0  23:0
		 */
		input = 1 << 16;
		pad = pcePadGetDirect();
		if(pad & PAD_A     ) input |= 1 << 0;
		if(pad & PAD_B     ) input |= 1 << 1;
		if(pad & PAD_SELECT) input |= 1 << 2;
		if(pad & PAD_START ) input |= 1 << 3;
		if(pad & PAD_UP    ) input |= 1 << 4;
		if(pad & PAD_DN    ) input |= 1 << 5;
		if(pad & PAD_LF    ) input |= 1 << 6;
		if(pad & PAD_RI    ) input |= 1 << 7;

		if(draw) {
			draw = NULL;
			ppu_reduce(vbuff, vbuff);
			schedule();
			if((joy & PAD_START) && (joy & PAD_SELECT)) {
				break;
			}
		}
	}

	if(sound) {
		stream_stop();
	}

	return 0;
}

